"""
Functions related to density and specific volume.

These are a subset of the TEOS-10 table category
"specific volume, density, and enthalpy".

We are grouping the functions related to enthalpy and internal energy
in their own "energy" module.
"""
from ._wrapped_ufuncs import (
    alpha,
    alpha_on_beta,
    beta,
    kappa,
    rho,
    rho_alpha_beta,
    rho_t_exact,
    sigma0,
    sigma1,
    sigma2,
    sigma3,
    sigma4,
    sound_speed,
    specvol,
    specvol_alpha_beta,
    specvol_anom_standard,
)



"""
Internally import from this, not from _wrapped_ufuncs.
Users should import only from non-private modules, of course.
"""

import numpy

from ._wrapped_ufuncs import *

_p_from_z = p_from_z
def p_from_z(z, lat, geo_strf_dyn_height=0, sea_surface_geopotential=0):
    return _p_from_z(z, lat, geo_strf_dyn_height, sea_surface_geopotential)
p_from_z.__doc__ = _p_from_z.__doc__

_z_from_p = z_from_p
def z_from_p(p, lat, geo_strf_dyn_height=0, sea_surface_geopotential=0):
    return _z_from_p(p, lat, geo_strf_dyn_height, sea_surface_geopotential)
z_from_p.__doc__ = _z_from_p.__doc__

_gibbs = gibbs
def gibbs(ns, nt, np, SA, t, p):
    params = {"ns": ns, "nt": nt, "np": np}
    for k, v in params.items():
        u = numpy.unique(v)
        if u.min() < 0 or u.max() > 2 or u.dtype.kind != "i":
            raise ValueError("ns, nt, np must contain integers 0, 1, or 2;"
                             f" found {k}={v}")
    return _gibbs(ns, nt, np, SA, t, p)
gibbs.__doc__ = _gibbs.__doc__


_gibbs_ice = gibbs_ice
def gibbs_ice(nt, np, t, p):
    params = {"nt": nt, "np": np}
    for k, v in params.items():
        u = numpy.unique(v)
        if u.min() < 0 or u.max() > 2 or u.dtype.kind != "i":
            raise ValueError("nt, np must contain integers 0, 1, or 2;"
                             f" found {k}={v}")
    return _gibbs_ice(nt, np, t, p)
gibbs_ice.__doc__ = _gibbs_ice.__doc__



"""
Conversions involving temperature, salinity, entropy, pressure,
and height.

Those most commonly used probably include:

- :func:`gsw.CT_from_t`
- :func:`gsw.SA_from_SP`
- :func:`gsw.SP_from_C`
- :func:`gsw.p_from_z`
- :func:`gsw.z_from_p`

"""

__all__ = ['t90_from_t68',
            'adiabatic_lapse_rate_from_CT',
            'C_from_SP',
            'CT_from_enthalpy',
            'CT_from_entropy',
            'CT_from_pt',
            'CT_from_rho',
            'CT_from_t',
            'deltaSA_from_SP',
            'entropy_from_pt',
            'entropy_from_t',
            'pt0_from_t',
            'pt_from_CT',
            'pt_from_entropy',
            'pt_from_t',
            'SA_from_rho',
            'SA_from_SP',
            'SA_from_Sstar',
            'SP_from_C',
            'SP_from_SA',
            'SP_from_SK',
            'SP_from_SR',
            'SP_from_Sstar',
            'SR_from_SP',
            'Sstar_from_SA',
            'Sstar_from_SP',
            't_from_CT',
            'p_from_z',
            'z_from_p',
            ]

from ._fixed_wrapped_ufuncs import (
    C_from_SP,
    CT_from_enthalpy,
    CT_from_entropy,
    CT_from_pt,
    CT_from_rho,
    CT_from_t,
    SA_from_rho,
    SA_from_SP,
    SA_from_Sstar,
    SP_from_C,
    SP_from_SA,
    SP_from_SK,
    SP_from_SR,
    SP_from_Sstar,
    SR_from_SP,
    Sstar_from_SA,
    Sstar_from_SP,
    adiabatic_lapse_rate_from_CT,
    deltaSA_from_SP,
    entropy_from_pt,
    entropy_from_t,
    p_from_z,
    pt0_from_t,
    pt_from_CT,
    pt_from_entropy,
    pt_from_t,
    t_from_CT,
    z_from_p,
)
from ._utilities import match_args_return


@match_args_return
def t90_from_t68(t68):
    """
    ITS-90 temperature from IPTS-68 temperature

    This conversion should be applied to all in-situ
    data collected between 1/1/1968 and 31/12/1989.

    """
    return t68 / 1.00024



"""
Functions not specific to the TEOS-10 realm of variables.
"""

import numpy as np

from . import _gsw_ufuncs
from ._utilities import indexer, match_args_return


@match_args_return
def pchip_interp(x, y, xi, axis=0):
    """
    Interpolate using Piecewise Cubic Hermite Interpolating Polynomial

    This is a shape-preserving algorithm; it does not introduce new local
    extrema.  The implementation in C that is wrapped here is largely taken
    from the scipy implementation,
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.PchipInterpolator.html.

    Points outside the range of the interpolation table are filled using the
    end values in the table.  (In contrast,
    scipy.interpolate.pchip_interpolate() extrapolates using the end
    polynomials.)

    Parameters
    ----------
    x, y : array-like
        Interpolation table x and y; n-dimensional, must be broadcastable to
        the same dimensions.
    xi : array-like
        One-dimensional array of new x values.
    axis : int, optional, default is 0
        Axis along which xi is taken.

    Returns
    -------
    yi : array
        Values of y interpolated to xi along the specified axis.

    """

    xi = np.array(xi, dtype=float, copy=False, order='C', ndmin=1)
    if xi.ndim > 1:
        raise ValueError('xi must be no more than 1-dimensional')
    nxi = xi.size
    x, y = np.broadcast_arrays(x, y)
    out_shape = list(x.shape)
    out_shape[axis] = nxi
    yi = np.empty(out_shape, dtype=float)
    yi.fill(np.nan)

    goodmask = ~(np.isnan(x) | np.isnan(y))

    order = 'F' if y.flags.fortran else 'C'
    for ind in indexer(y.shape, axis, order=order):
        igood = goodmask[ind]
        # If p_ref is below the deepest value, skip the profile.
        xgood = x[ind][igood]
        ygood = y[ind][igood]

        yi[ind] = _gsw_ufuncs.util_pchip_interp(xgood, ygood, xi)

    return yi



import os

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal, assert_array_equal

import gsw
from gsw._utilities import Bunch

root_path = os.path.abspath(os.path.dirname(__file__))

cv = Bunch(np.load(os.path.join(root_path, 'gsw_cv_v3_0.npz')))
# Override the original with what we calculate using pchip interp
cv.geo_strf_velocity = np.load(os.path.join(root_path,'geo_strf_velocity.npy'))

lon = [1, 2]
lat = [45, 45]
expected = 78626.18767687

# distance tests

def test_list():
    value = gsw.distance(lon, lat, p=0, axis=-1)
    assert_almost_equal(expected, value)

def test_1darray():
    value = gsw.distance(np.array(lon), np.array(lat), p=0, axis=-1)
    assert_almost_equal(expected, value)

def test_1darray_default_p():
    # @match_args_return doesn't see the default p.
    value = gsw.distance(np.array(lon), np.array(lat))
    assert_almost_equal(expected, value)

def test_2dlist():
    value = gsw.distance(np.atleast_2d(lon), np.atleast_2d(lat), p=0, axis=1)
    assert_almost_equal(expected, value)

# geostrophic streamfunction tests

def test_strf_no_good():
    # revised geo_strf_dyn_height requires 2 valid points
    shape = (5,)
    SA = np.ma.masked_all(shape, dtype=float)
    CT = np.ma.masked_all(shape, dtype=float)
    p = np.array([0.0, 10.0, 20.0, 30.0, 40.0])

    # No valid points.
    strf = gsw.geo_strf_dyn_height(SA, CT, p, p_ref=0, axis=0)
    expected = np.zeros(shape, float) + np.nan
    assert_array_equal(strf.filled(np.nan), expected)

    # 1 valid point: not enough.
    SA[:1] = 35
    CT[:1] = 5
    strf = gsw.geo_strf_dyn_height(SA, CT, p, p_ref=0, axis=0)
    expected = np.zeros(shape, float) + np.nan
    assert_array_equal(strf.filled(np.nan), expected)

    # 2 valid points: enough for the calculation to proceed.
    SA[:2] = 35
    CT[:2] = [5, 4]
    strf = gsw.geo_strf_dyn_height(SA, CT, p, p_ref=0, axis=0)
    assert strf.count() == 2

def test_geostrophy():
    lon = cv.long_chck_cast
    lat = cv.lat_chck_cast
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    SA = cv.SA_chck_cast
    strf = gsw.geo_strf_dyn_height(SA, CT, p)
    geovel, midlon, midlat = gsw.geostrophic_velocity(strf, lon, lat)
    assert_almost_equal(geovel, cv.geo_strf_velocity)
    assert_almost_equal(midlon, cv.geo_strf_velocity_mid_long[0])
    assert_almost_equal(midlat, cv.geo_strf_velocity_mid_lat[0])

def test_dyn_height_shallower_pref():
    """
    Check that we can handle a p_ref that is shallower than the top of the
    cast.  To make the results from bin 1 on down independent of whether
    bin 0 has been deleted, we need to use linear interpolation.
    """
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    SA = cv.SA_chck_cast
    strf0 = gsw.geo_strf_dyn_height(SA, CT, p, p_ref=0, interp_method='linear')
    strf1 = gsw.geo_strf_dyn_height(SA[1:], CT[1:], p[1:], p_ref=0,
                                    interp_method='linear')
    found = strf1 - strf1[0]
    expected = strf0[1:] - strf0[1]
    assert_almost_equal(found, expected)

def test_dyn_height_grid_bug():
    """
    The C toolbox had a bug in the interpolation grid generation, triggered
    under specific circumstances demonstrated in issue #225.  The fix was copied
    from https://github.com/TEOS-10/GSW-C/pull/87.
    """
    SA = np.array([34.17634373, 34.17633109, 34.17725485, 34.90704565])
    CT = np.array([ 0.02005831,  0.02003146,  0.01868686, 1.7676452 ])
    pres = np.array([  4.9000001 ,   5.69999981,  15.80000019, 995.90002441])
    # Last 2 args: grid delta-p is 1; interp method is 3 (mrst).
    dh1 = gsw._gsw_ufuncs.geo_strf_dyn_height_1(SA, CT, pres, pres[-1], 1, 3)
    assert dh1[-1] == 0
    dh2 = gsw.geo_strf_dyn_height(SA, CT, pres, p_ref = pres[-1], interp_method="mrst")
    assert_array_equal(dh1, dh2)

def test_pz_roundtrip():
    """
    The p_z conversion functions have Matlab-based checks that use
    only the first two arguments.
    Here we verify that the functions are also inverses when the optional
    arguments are used.
    """
    z = np.array([-10, -100, -1000, -5000], dtype=float)
    p = gsw.p_from_z(z, 30, 0.5, 0.25)
    zz = gsw.z_from_p(p, 30, 0.5, 0.25)
    assert_almost_equal(z, zz)

def test_dyn_height_mrst():
    """
    Tests the MRST-PCHIP interpolation method.
    """
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    SA = cv.SA_chck_cast
    pr = cv.pr
    strf = gsw.geo_strf_dyn_height(SA, CT, p, p_ref=pr, interp_method='mrst')

    assert_allclose(strf, cv.geo_strf_dyn_height, rtol=0, atol=cv.geo_strf_dyn_height_ca)

def test_steric_height_mrst():
    """
    Tests the MRST-PCHIP interpolation method.
    """
    p = cv.p_chck_cast
    CT = cv.CT_chck_cast
    SA = cv.SA_chck_cast
    pr = cv.pr
    strf = gsw.geo_strf_steric_height(SA, CT, p, p_ref=pr, interp_method='mrst')

    assert_allclose(strf, cv.geo_strf_steric_height, rtol=0, atol=cv.geo_strf_steric_height_ca)



import numpy as np
import pytest

import gsw

nx, ny, nz = 2, 3, 10
y = np.arange(nx*ny*nz, dtype=float).reshape((nx, ny, nz))
y += y**1.5
z = np.arange(nz, dtype=float)
z = np.broadcast_to(z, y.shape)
zn = z.copy()
zn[:, :, [0, -1]] = np.nan

xi_arraylist = [[0.5, 1.5], np.linspace(-1, z.max() + 10, 50)]

# Initial smoke test with small and large xi arrays.
@pytest.mark.parametrize("xi", xi_arraylist)
def test_in_range(xi):
    yi = gsw.pchip_interp(z, y, xi, axis=-1)
    assert yi.shape == (nx, ny, len(xi))


# Try with NaNs.
@pytest.mark.parametrize("xi", xi_arraylist)
def test_in_range_nan(xi):
    yi = gsw.pchip_interp(zn, y, xi, axis=-1)
    assert yi.shape == (nx, ny, len(xi))



"""
Generate our own test values for geostrophic calculations.

We are presently using simple pchip interpolation for
geostrophy rather than trying to mimic the ever-changing
Matlab functions. Therefore we have to make our own test
values, using the current test cast inputs.

This is a minimal script for that purpose, to be run in
the tests directory in which it lives.  It should be run
only if we change to a different calculation algorithm,
or we update the cast input and general check value file.
"""

import numpy as np

import gsw
from gsw._utilities import Bunch

cv = Bunch(np.load('gsw_cv_v3_0.npz'))

dyn_height = gsw.geo_strf_dyn_height(cv.SA_chck_cast,
                                     cv.CT_chck_cast,
                                     cv.p_chck_cast,
                                     cv.pr)
np.save('geo_strf_dyn_height.npy', dyn_height)

steric_height = gsw.geo_strf_steric_height(cv.SA_chck_cast,
                                     cv.CT_chck_cast,
                                     cv.p_chck_cast,
                                     cv.pr)
np.save('geo_strf_steric_height.npy', steric_height)

lon = cv.long_chck_cast
lat = cv.lat_chck_cast
p = cv.p_chck_cast
CT = cv.CT_chck_cast
SA = cv.SA_chck_cast
strf = gsw.geo_strf_dyn_height(SA, CT, p)
geovel, midlon, midlat = gsw.geostrophic_velocity(strf, lon, lat)
np.save('geo_strf_velocity.npy', geovel)
# midlon, midlat are OK



"""
Freezing-point functions.
"""

from ._wrapped_ufuncs import (
    CT_freezing,
    CT_freezing_first_derivatives,
    CT_freezing_first_derivatives_poly,
    CT_freezing_poly,
    SA_freezing_from_CT,
    SA_freezing_from_CT_poly,
    SA_freezing_from_t,
    SA_freezing_from_t_poly,
    pot_enthalpy_ice_freezing,
    pot_enthalpy_ice_freezing_first_derivatives,
    pot_enthalpy_ice_freezing_first_derivatives_poly,
    pot_enthalpy_ice_freezing_poly,
    pressure_freezing_CT,
    t_freezing,
    t_freezing_first_derivatives,
    t_freezing_first_derivatives_poly,
)
